#pragma once

#include <print>
#include <string>
#include <unordered_map>
#include <vector>

class ArgParser {
public:
  void addArg(const std::string& shortOpt, const std::string& longOpt, const std::string& description, bool hasValue = true) {
    m_arguments.emplace_back(shortOpt, longOpt, description, "", hasValue);
    if (!shortOpt.empty()) {
      m_shortOptMap[shortOpt] = &m_arguments.back();
    }
    if (!longOpt.empty()) {
      m_longOptMap[longOpt] = &m_arguments.back();
    }
  }

  bool parse(const int argc, char* argv[]) {
    for (int i{1}; i < argc; ++i) {
      if (const std::string arg{argv[i]}; m_shortOptMap.contains(arg)) {
        if (arg == "--help" || arg == "-h") {
          printHelp();
          return false;
        }
        m_shortOptMap[arg]->exists = true;
        if (m_shortOptMap[arg]->hasValue && i + 1 < argc) {
          m_shortOptMap[arg]->value = argv[++i];
        }
      } else if (m_longOptMap.contains(arg)) {
        if (arg == "--help" || arg == "-h") {
          printHelp();
          return false;
        }
        m_longOptMap[arg]->exists = true;
        if (m_longOptMap[arg]->hasValue && i + 1 < argc) {
          m_longOptMap[arg]->value = argv[++i];
        }
      } else {
        throw std::invalid_argument("Unknown option: " + arg);
      }
    }
    return true;
  }

  void printHelp() const {
    std::println("Usage: [options]");
    for (const auto& arg : m_arguments) {
      std::println("   {}, {}: {}", arg.shortOpt, arg.longOpt, arg.description);
    }
  }

  template <typename T>
  T get(const std::string& arg) const {
    std::string value;
    if (m_longOptMap.contains(arg)) {
      value = m_longOptMap.at(arg)->value;
    } else if (m_shortOptMap.contains(arg)) {
      value = m_shortOptMap.at(arg)->value;
    } else {
      throw std::runtime_error(arg + " has no value");
    }
    if constexpr (std::is_same_v<T, int>) {
      return std::stoi(value);
    } else if constexpr (std::is_same_v<T, float>) {
      return std::stof(value);
    } else if constexpr (std::is_same_v<T, double>) {
      return std::stod(value);
    } else {
      return value;
    }
  }

  bool has(const std::string& arg) const {
    if (const auto longOptIt = m_longOptMap.find(arg); longOptIt != m_longOptMap.end()) {
      return longOptIt->second->exists;
    }
    if (const auto shortOptIt = m_shortOptMap.find(arg); shortOptIt != m_shortOptMap.end()) {
      return shortOptIt->second->exists;
    }
    return false;
  }

private:
  struct Argument {
    std::string shortOpt;
    std::string longOpt;
    std::string description;
    std::string value;
    bool        hasValue = true;
    bool        exists   = false;
  };

  std::vector<Argument>                      m_arguments{};
  std::unordered_map<std::string, Argument*> m_shortOptMap{};
  std::unordered_map<std::string, Argument*> m_longOptMap{};
};
